/* ***** BEGIN LICENSE BLOCK *****
 * Version: AGPL 3.0
 *
 * The contents of this file may be used under the terms of the
 * Affero GNU General Public License Version 3 or later (the "AGPL").
 *
 * ***** END LICENSE BLOCK ***** */

var soap = new Namespace("soap", "http://schemas.xmlsoap.org/soap/envelope/");
var t = new Namespace("t", "http://schemas.microsoft.com/exchange/services/2006/types");
var m = new Namespace("m", "http://schemas.microsoft.com/exchange/services/2006/messages");

const participationMap = {
	"Unknown"	: "NEEDS-ACTION",
	"Tentative"	: "TENTATIVE",
	"Accept"	: "ACCEPTED",
	"Decline"	: "DECLINED",
	"Organizer"	: "ACCEPTED"
};

const dayMap = {
	'Monday'	: 'MO',
	'Tuesday'	: 'TU',
	'Wednesday'	: 'WE',
	'Thursday'	: 'TH',
	'Friday'	: 'FR',
	'Saturday'	: 'SA',
	'Sunday'	: 'SU',
	'Weekday'	: ['MO', 'TU', 'WE', 'TH', 'FR'],
	'WeekendDay'	: ['SA', 'SO'],
	'Day'		: ['MO', 'TU', 'WE', 'TH', 'FR', 'SA', 'SO']
};

const weekMap = {
	'First'		: 1,
	'Second'	: 2,
	'Third'		: 3,
	'Fourth'	: 4,
	'Last'		: -1
};

const monthMap = {
	'January'	: 1,
	'February'	: 2,
	'March'		: 3,
	'April'		: 4,
	'May'		: 5,
	'June'		: 6,
	'July'		: 7,
	'August'	: 8,
	'September'	: 9,
	'October'	: 10,
	'November'	: 11,
	'December'	: 12
};


function readRecurrenceRule(recit)
{
	/*
	 * The Mozilla recurrence API is bothersome and dictated by libical.
	 *
	 * We need to obey and build an iCalendar string which we feed in
	 * to get the proper recurrence info.
	 */

	var comps = {};

	for each (var rec in recit) {
		switch (rec.localName()) {
		case "RelativeYearlyRecurrence":
		case "AbsoluteYearlyRecurrence":
			comps['FREQ'] = "YEARLY";
			break;
		case "RelativeMonthlyRecurrence":
		case "AbsoluteMonthlyRecurrence":
			comps['FREQ'] = "MONTHLY";
			break;
		case "WeeklyRecurrence":
			comps['FREQ'] = "WEEKLY";
			break;
		case "DailyRecurrence":
			comps['FREQ'] = "DAILY";
			break;
		case "NoEndRecurrence":
		case "EndDateRecurrence":
		case "NumberedRecurrence":
			break;
		default:
			sdbg("skipping " + rec.localName() + "\n");
			continue;
		}

		var weekdays = [];
		var week = [];
		for each (var comp in rec) {
			switch (comp.localName()) {
			case 'DaysOfWeek':
				for each (let day in comp.toString().split(" ")) {
					weekdays = weekdays.concat(dayMap[day]);
				}
				break;
			case 'DayOfWeekIndex':
				week = weekMap[comp.toString()];
				break;
			case 'Month':
				comps['BYMONTH'] = monthMap[comp.toString()];
				break;
			case 'DayOfMonth':
				comps['BYMONTHDAY'] = comp.toString();
				break;
			case 'FirstDayOfWeek':
				comps['WKST'] = dayMap[comp.toString()];
				break;
			case 'Interval':
				comps['INTERVAL'] = comp.toString();
				break;
			case 'StartDate':
				/* Dunno what to do with this; no place to set */
				break;
			case 'EndDate':
				comps['UNTIL'] = comp.toString().replace(/Z$/, '');
				break;
			case 'NumberOfOccurrences':
				comps['COUNT'] = comp.toString();
				break;
			}
		}

		let wdtemp = weekdays;
		weekdays = [];
		for each (let day in wdtemp) {
			weekdays.push(week + day);
		}
		if (weekdays.length > 0) {
			comps['BYDAY'] = weekdays.join(',');
		}
	}

	var compstrs = [];
	for (var k in comps) {
		compstrs.push(k + "=" + comps[k]);
	}

	if (compstrs.length == 0) {
		return null;
	}

	var recrule = cal.createRecurrenceRule();

	/* need to do this re-assign game so that the string gets parsed */
	var prop = recrule.icalProperty;
	prop.value = compstrs.join(';');
	recrule.icalProperty = prop;

	return recrule;
}

function readRecurrence(cit, it)
{
	var recrule = readRecurrenceRule(it.t::Recurrence.children());

	if (recrule === null) {
		return null;
	}

	var recurrenceInfo = cal.createRecurrenceInfo(cit);
	recurrenceInfo.setRecurrenceItems(1, [recrule]);

	return recurrenceInfo;
}

function readModifiedOccurrences(cit, it, exitems)
{
	for each (var mit in it.t::ModifiedOccurrences.children()) {
		var key = [mit.t::ItemId.@Id, mit.t::ItemId.@ChangeKey];
		var oit = exitems[key];

		if (!oit || !cit.recurrenceInfo) {
			continue;
		}
		/*
		 * We need to clear alarmLastAck, because it has been
		 * read before and might contain some value we need
		 * to ignore.  All alarm dismiss/snooze is handled
		 * in the parent item.
		 */
		oit.alarmLastAck = null;
		cit.recurrenceInfo.modifyException(oit, true);
		delete exitems[key];
	}
}

function readDeletedOccurrences(cit, it)
{
	for each (var mit in it.t::DeletedOccurrences.children()) {
		cit.recurrenceInfo.removeOccurrenceAt(cal.fromRFC3339(mit.t::Start, cal.UTC()));
	}
}

function formatTZoff(off)
{
	var s = "";

	/*
	 * For some very special reason EWA and RFC5545 use different
	 * opinions on what "UTC offset" and "offset from UTC" means.
	 */
	if (off.isNegative) {
		s += "+";
	} else {
		s += "-";
	}
	if (off.hours < 10) {
		s += "0";
	}
	s += off.hours;
	if (off.minutes < 10) {
		s += "0";
	}
	s += off.minutes;
	s += "\n";

	return s;
}

function formatTZ(tztype, tzstart, tzrr, tzofffrom, tzoffto)
{
	var icaltz = "";

	icaltz += "BEGIN:" + tztype + "\n";
	icaltz += "DTSTART:" + tzstart + "\n";
	if (tzrr) {
		icaltz += tzrr.icalProperty;
	}
	icaltz += "TZOFFSETFROM:" + formatTZoff(tzofffrom);
	icaltz += "TZOFFSETTO:" + formatTZoff(tzoffto);
	icaltz += "END:" + tztype + "\n";

	return icaltz;
}

function readMeetingTimezone(cit, it)
{
	/*
	 * Timezone handling is even more iCalendar centric than recurrence
	 * handling.  We have to mock up an ical string for the timezone,
	 * then pack it into a VCALENDAR, parse the whole mess, extract back
	 * the VTIMEZONE component, and mock up a calITimezone object with the
	 * parsed icalComponent.
	 */
	var tz = it.t::MeetingTimeZone[0];

	if (!tz) {
		return;
	}

	var tzname = tz.@TimeZoneName.toString();

	var baseoff = cal.createDuration(tz.t::BaseOffset.toString());
	var dstoff = baseoff;
	var stdtoff = baseoff;

	var stdttimestr = "19700101T000000";

	var dsttz = tz.t::Daylight[0];
	var stdttz = tz.t::Standard[0];

	if (dsttz && stdttz) {
		dstoff = cal.createDuration(dsttz.t::Offset);
		dstoff.addDuration(baseoff);
		var dstrr = readRecurrenceRule(dsttz.t::RelativeYearlyRecurrence);
		var dsttimestr = '19700101T' + dsttz.t::Time.toString().replace(/:/g, '');

		stdtoff = cal.createDuration(stdttz.t::Offset);
		stdtoff.addDuration(baseoff);
		var stdtrr = readRecurrenceRule(stdttz.t::RelativeYearlyRecurrence);
		stdttimestr = '19700101T' + stdttz.t::Time.toString().replace(/:/g, '');
	}

	var icaltz = "";

	icaltz += "BEGIN:VCALENDAR\n";
	icaltz += "BEGIN:VTIMEZONE\n";
	icaltz += "TZID:" + tzname + "\n";
	if (dsttz) {
		icaltz += formatTZ('DAYLIGHT', dsttimestr, dstrr, stdtoff, dstoff);
	}
	icaltz += formatTZ('STANDARD', stdttimestr, stdtrr, dstoff, stdtoff);
	icaltz += "END:VTIMEZONE\n";
	icaltz += "END:VCALENDAR\n";

	var calcomp = cal.getIcsService().parseICS(icaltz, null);
	var tzcomp = calcomp.getFirstSubcomponent("VTIMEZONE");
	var ctz = {
		//provider: cprov,
		icalComponent: tzcomp,
		tzid: tzname,
		isFloating: false,
		isUTC: false,
		displayName: tzname,
		toString: function() {
			return tzname;
		}
	};

	if (dstoff.inSeconds == 0 && stdtoff.inSeconds == 0) {
		ctz = cal.UTC();
	}

	cit.startDate = cit.startDate.getInTimezone(ctz);
	cit.endDate = cit.endDate.getInTimezone(ctz);
}

function readXmlAttendee(at, type)
{
	let mbox = at.t::Mailbox;
	let attendee = createAttendee();

	if (!type) {
		type = "REQ-PARTICIPANT";
	}

	attendee.id = 'mailto:' + mbox.t::EmailAddress.toString();
	attendee.commonName = mbox.t::Name.toString();
	attendee.rsvp = "FALSE";
	attendee.userType = "INDIVIDUAL";
	attendee.role = type;

	if (at.t::ResponseType.length() > 0) {
		attendee.participationStatus = participationMap[at.t::ResponseType.toString()];
	}

	return attendee;
}

function readDismissSnoozeState(aItem, it)
{
	let nextRem = it.t::ExtendedProperty.(t::ExtendedFieldURI.@PropertyId == '34144').t::Value[0];

	if (!nextRem) {
		return;
	}

	nextRem = cal.fromRFC3339(nextRem.toString(), cal.UTC());
	let lastAck = nextRem.clone();
	lastAck.addDuration(cal.createDuration('-PT1S'));

	aItem.alarmLastAck = lastAck;

	if (nextRem.compare(cal.fromRFC3339("4501-01-01T00:00:00Z", cal.UTC())) == 0) {
		return;
	}

	if (!aItem.recurrenceInfo) {
		aItem.setProperty('X-MOZ-SNOOZE-TIME', nextRem.icalString);
	} else {
		let nextOcc = aItem.recurrenceInfo.getNextOccurrence(lastAck);
		let alarm = aItem.getAlarms({})[0];

		if (nextOcc && alarm) {
			let nextOccAlarm = cal.alarms.calculateAlarmDate(nextOcc, alarm);
			let alarmDiff = nextOccAlarm.compare(nextRem);

			if (alarmDiff == 0) {
				/* no item was snoozed, this is the regular alarm date */
			} else {
				if (alarmDiff > 0) {
					/*
					 * The alarm of the next item will happen after
					 * the remider time.  This means we did snooze
					 * a previous alarm.
					 */
					nextOcc = aItem.recurrenceInfo.getPreviousOccurrence(nextRem);
				}
				aItem.setProperty('X-MOZ-SNOOZE-TIME-' + nextOcc.recurrenceId.nativeTime,
						  nextRem.icalString);
			}
		}
	}
}

function readXmlItem(aCal, it, exitems)
{
	var cit = cal.createEvent();
	cit.calendar = aCal.superCalendar;

	cit.id = it.t::UID.toString();
	cit.setProperty("X-ITEMID", it.t::ItemId.@Id.toString());
	cit.setProperty("SEQUENCE", it.t::ItemId.@ChangeKey.toString());

	cit.title = it.t::Subject;
	cit.setProperty("DESCRIPTION", it.t::Body.(@BodyType == "Text").toString());

	var cats = [];
	for each (var cat in it.t::Categories.t::String) {
		cats.push(cat.toString());
	}
	cit.setCategories(cats.length, cats);

	if (it.t::ReminderIsSet == "true") {
		let alarmOffset = cal.createDuration();
		alarmOffset.minutes = - it.t::ReminderMinutesBeforeStart;
		alarmOffset.normalize();

		let alarm = cal.createAlarm();
		alarm.action = "DISPLAY";
		alarm.related = Components.interfaces.calIAlarm.ALARM_RELATED_START;
		alarm.offset = alarmOffset;

		cit.addAlarm(alarm);
	}

	var defaultTz = cal.calendarDefaultTimezone();
	cit.startDate = cal.fromRFC3339(it.t::Start, defaultTz).getInTimezone(defaultTz);
	cit.endDate = cal.fromRFC3339(it.t::End, defaultTz).getInTimezone(defaultTz);

	if (it.t::IsAllDayEvent == "true") {
		cit.startDate.isDate = true;
		cit.endDate.isDate = true;
	}

	cit.setProperty("DTSTAMP", _f["fromRFC3339"](it.t::DateTimeReceived, cal.UTC()));
	cit.setProperty("LOCATION", it.t::Location.toString());

	if (it.t::Organizer.length() > 0) {
		let org = readXmlAttendee(it.t::Organizer);
		org.isOrganizer = true;
		cit.organizer = org;
	}

	for each (var at in it.t::RequiredAttendees.t::Attendee) {
		cit.addAttendee(readXmlAttendee(at), "REQ-PARTICIPANT");
	}
	for each (var at in it.t::OptionalAttendees.t::Attendee) {
		cit.addAttendee(readXmlAttendee(at), "OPT-PARTICIPANT");
	}

	var me = aCal.getInvitedAttendee(cit);
	if (me) {
		var stat = it.t::MyResponseType.toString() || "Unknown";

		me.participationStatus = participationMap[stat];
	}

	cit.recurrenceId = _f["fromRFC3339"](it.t::RecurrenceId, cal.UTC());
	cit.recurrenceInfo = readRecurrence(cit, it);
	readModifiedOccurrences(cit, it, exitems);
	readDeletedOccurrences(cit, it);

	readMeetingTimezone(cit, it);

	readDismissSnoozeState(cit, it);

	return cit;
}

function readGetItemReply(aCal, aRep, aItemFilter, aRangeStart, aRangeEnd)
{
	var rm = aRep..m::GetItemResponseMessage;

	var searchtype;
	if (aItemFilter & Components.interfaces.calICalendar.ITEM_FILTER_REQUEST_NEEDS_ACTION) {
		searchtype = 'MeetingRequest';
	} else {
		searchtype = 'CalendarItem';
	}

	var exitems = [];
	for each (var it in rm.m::Items.children(searchtype).(t::CalendarItemType == "Exception")) {
		var citem = readXmlItem(aCal, it);
		exitems[[it.t::ItemId.@Id, it.t::ItemId.@ChangeKey]] = citem;
	}

	var items = [];
	for each (var it in rm.m::Items.children(searchtype).(t::CalendarItemType != "Exception")) {
		var citem = readXmlItem(aCal, it, exitems);

		if (aItemFilter & Components.interfaces.calICalendar.ITEM_FILTER_CLASS_OCCURRENCES) {
			var expanditems = citem.getOccurrencesBetween(aRangeStart, aRangeEnd, {});
			expanditems = expanditems || [citem];
			items = items.concat(expanditems);
		} else {
			items.push(citem);
		}
	}

	return items;
}

function readFindItemReply(aRep)
{
	/*
	 * We want to include all Single items, all Exception items, but also
	 * at least one Occurrence or Exception item for each master.
	 * If we include too many Occurrences, we will query for the master
	 * too often, but if we don't include any, we might not query for the
	 * master at all.
	 *
	 * We first collect all non-Occurrences, and after that we fill in
	 * Occurrence for those masters that did not yet see any Exception.
	 */

	var uids = {};
	var ids = [];
	for each (var e in aRep..t::Items.*.(t::CalendarItemType != "Occurrence")) {
		ids.push({Id: e.t::ItemId.@Id.toString(),
			  ChangeKey: e.t::ItemId.@ChangeKey.toString(),
			  type: e.t::CalendarItemType.toString(),
			  uid: e.t::UID.toString()});
		uids[e.t::UID.toString()] = true;
	}

	for each (var e in aRep..t::Items.*.(t::CalendarItemType == "Occurrence")) {
		if (uids[e.t::UID.toString()]) {
			continue;
		}
		ids.push({Id: e.t::ItemId.@Id.toString(),
			  ChangeKey: e.t::ItemId.@ChangeKey.toString(),
			  type: e.t::CalendarItemType.toString(),
			  uid: e.t::UID.toString()});
		uids[e.t::UID.toString()] = true;
	}

	return ids;
}
