/* ***** BEGIN LICENSE BLOCK *****
 * Version: AGPL 3.0
 *
 * The contents of this file may be used under the terms of the
 * Affero GNU General Public License Version 3 or later (the "AGPL").
 *
 * ***** END LICENSE BLOCK ***** */

var xml_tag = '<?xml version="1.0" encoding="utf-8"?>\n';
var soap = new Namespace("soap", "http://schemas.xmlsoap.org/soap/envelope/");
var t = new Namespace("t", "http://schemas.microsoft.com/exchange/services/2006/types");
var m = new Namespace("m", "http://schemas.microsoft.com/exchange/services/2006/messages");

const fieldPathMap = {
	'Body'			: 'item',
	'CalendarItemType'	: 'calendar',
	'Categories'		: 'item',
	'DateTimeReceived'	: 'item',
	'DeletedOccurrences'	: 'calendar',
	'End'			: 'calendar',
	'FirstOccurrence'	: 'calendar',
	'IsAllDayEvent'		: 'calendar',
	'IsCancelled'		: 'calendar',
	'LastOccurrence'	: 'calendar',
	'Location'		: 'calendar',
	'MeetingTimeZone'	: 'calendar',
	'ModifiedOccurrences'	: 'calendar',
	'MyResponseType'	: 'calendar',
	'OptionalAttendees'	: 'calendar',
	'Organizer'		: 'calendar',
	'Recurrence'		: 'calendar',
	'RecurrenceId'		: 'calendar',
	'ReminderDueBy'		: 'item',
	'ReminderIsSet'		: 'item',
	'ReminderMinutesBeforeStart' : 'item',
	'RequiredAttendees'	: 'calendar',
	'Start'			: 'calendar',
	'Subject'		: 'item',
	'UID'			: 'calendar',
};

const dayRevMap = {
	'MO' : 'Monday',
	'TU' : 'Tuesday',
	'WE' : 'Wednesday',
	'TH' : 'Thursday',
	'FR' : 'Friday',
	'SA' : 'Saturday',
	'SU' : 'Sunday'
};

const dayIdxMap = ['SU', 'MO', 'TU', 'WE', 'TH', 'FR', 'SA'];

const weekRevMap = {
	'1' : 'First',
	'2' : 'Second',
	'3' : 'Third',
	'4' : 'Fourth',
	'-1': 'Last'
};

const monthIdxMap = ['January', 'February', 'March', 'April', 'May', 'June',
		     'July', 'August', 'September', 'October', 'November', 'December'];


function makeAutodiscover(aEmail)
{
	var req = <Autodiscover xmlns="http://schemas.microsoft.com/exchange/autodiscover/outlook/requestschema/2006"/>;

	req.Request.EMailAddress = aEmail;
	req.Request.AcceptableResponseSchema = "http://schemas.microsoft.com/exchange/autodiscover/outlook/responseschema/2006a";

	return xml_tag + String(req);
}

function makeSoapMessage(aReq)
{
	var msg = <soap:Envelope xmlns:soap={soap} xmlns:t={t} xmlns:m={m}/>;

	msg.soap::Header.t::RequestServerVersion.@Version = "Exchange2007_SP1";
	msg.soap::Body.request = aReq;

	return xml_tag + String(msg);
}

function makeFindItem(aItemFilter, aCount, aRangeStart, aRangeEnd)
{
	var folder = "calendar";

	var req = <m:FindItem xmlns:m={m} xmlns:t={t}/>;
	req.@Traversal = "Shallow";

	req.m::ItemShape.t::BaseShape = "IdOnly";
	req.m::ItemShape.t::AdditionalProperties.content = <>
	    <t:FieldURI FieldURI="calendar:UID" xmlns:t={t}/>
	    <t:FieldURI FieldURI="calendar:CalendarItemType" xmlns:t={t}/>
	    </>;

	// no ranges if searching for invitations XXX?
	var wantInvitations = (aItemFilter & Components.interfaces.calICalendar.ITEM_FILTER_REQUEST_NEEDS_ACTION) != 0;
	if (wantInvitations) {
		folder = "inbox";

		var restr =
			<m:Restriction xmlns:m={m} xmlns:t={t}>
				<t:IsEqualTo>
					<t:FieldURI FieldURI="item:ItemClass"/>
					<t:FieldURIOrConstant>
						<t:Constant Value="IPM.Schedule.Meeting.Request"/>
					</t:FieldURIOrConstant>
				</t:IsEqualTo>
			</m:Restriction>;

		req.appendChild(restr);
	} else if (aRangeStart && aRangeEnd) {
		var view = <m:CalendarView xmlns:m={m}/>;
		view.@StartDate = convDate(aRangeStart);
		view.@EndDate = convDate(aRangeEnd);

		if (aCount != 0) {
			view.@MaxEntriesReturned = aCount;
		}

		req.appendChild(view);
	}

	req.m::ParentFolderIds.t::DistinguishedFolderId.@Id = folder;

	return makeSoapMessage(req);
}

function hex2b64(hexstr)
{
	let binstr = '';

	for each (let hexchar in hexstr.match(/../g)) {
		let binchar = String.fromCharCode('0x'+hexchar);
		binstr += binchar;
	}

	return binstr;
}

function makeGlobalIdRestriction(aUID)
{
	/*
	 * We can not use FindItem with an UID directly.  Instead, we have to
	 * convert the UID to a GlobalObjectId, since the UID is generated from
	 * the GlobalObjectId.  There seem to be two different ways the UID is
	 * generated:
	 *
	 * 1) Same binary data: UID is encoded in hex, GID is in base64.
	 *    We just re-code the data.
	 *    This seems to be the case for all items created where Exchange
	 *    generates the GID itself.
	 *
	 * 2) Embedded arbitrary data: the UID is embedded in the GID.
	 *    We need to re-wrap the UID into the MAPI format that Exchange uses.
	 *    <http://msdn.microsoft.com/en-us/library/ee157690(v=EXCHG.80).aspx>
	 *    Not documented in this URL is that the unique data is the external
	 *    UID, NUL-terminated and prefixed with "vCal-Uid" and a
	 *    little endian 32 bit integer that always was 1 in my tests.
	 *    This is probably another counter to ensure uniqueness, but for now
	 *    we just assume for it to be 1.
	 *    The resulting string needs to be base64 encoded.
	 *
	 * We create a Restriction that matches (1) or (2), i.e. we search for both
	 * versions.
	 */
	let gidprefix = '040000008200e00074c5b7101a82e0080000000000000000000000000000000000000000';
	let vcaluid = 'vCal-Uid\1\0\0\0' + aUID + '\0';
	let gid1 = btoa(hex2b64(aUID));

	let gid2 = hex2b64(gidprefix);
	for (let i = 0; i < 4; i++) {
		gid2 += String.fromCharCode((vcaluid.length >> (i * 8)) % 256);
	}
	gid2 += vcaluid;
	gid2 = btoa(gid2);

	let r1 = <t:IsEqualTo xmlns:t={t}>
			<t:ExtendedFieldURI DistinguishedPropertySetId="Meeting" PropertyId="3" PropertyType="Binary" xmlns:t={t}/>
			<t:FieldURIOrConstant>
				<t:Constant Value={gid1}/>
			</t:FieldURIOrConstant>
		</t:IsEqualTo>;

	let r2 = r1.copy();
	r2..t::Constant.@Value = gid2;

	let restr = <t:Or xmlns:t={t}/>;
	restr.appendChild(r1);
	restr.appendChild(r2);

	return restr;
}

function makeFindItemUID(aUID)
{
	var req = <m:FindItem xmlns:m={m} xmlns:t={t}/>;
	req.@Traversal = "Shallow";

	req.m::ItemShape.t::BaseShape = "IdOnly";
	req.m::ItemShape.t::AdditionalProperties.content = <>
	    <t:FieldURI FieldURI="calendar:UID" xmlns:t={t}/>
	    <t:FieldURI FieldURI="calendar:CalendarItemType" xmlns:t={t}/>
	    </>;

	req.m::Restriction.content = makeGlobalIdRestriction(aUID);

	req.m::ParentFolderIds.t::DistinguishedFolderId.@Id = "calendar";

	return makeSoapMessage(req);

	// XXX missing: set organizer (use extended prop?)
}

function makeGetItemId(aItem)
{
	var req = <m:GetItem xmlns:m={m} xmlns:t={t}/>;

	req.m::ItemShape.t::BaseShape = 'IdOnly';
	req.m::ItemShape.t::BodyType = 'Text';
	req.m::ItemIds.t::ItemId.@Id = aItem.getProperty("X-ITEMID");

	return makeSoapMessage(req);
}

function makeGetItem(aIds, aItemFilter)
{
	var req = <m:GetItem xmlns:m={m} xmlns:t={t}/>;

	req.m::ItemShape.t::BaseShape = 'IdOnly';
	req.m::ItemShape.t::BodyType = 'Text';

	var aprop = new XMLList();

	for (var f in fieldPathMap) {
		var p = fieldPathMap[f] + ':' + f;

		aprop += <t:FieldURI FieldURI={p} xmlns:t={t}/>;
	}

	// snoozed until
	aprop += <t:ExtendedFieldURI DistinguishedPropertySetId="Common" PropertyId="34144" PropertyType="SystemTime" xmlns:t={t}/>;

	// global object id
	aprop += <t:ExtendedFieldURI DistinguishedPropertySetId="Meeting" PropertyId="3" PropertyType="Binary" xmlns:t={t}/>;

	req.m::ItemShape.t::AdditionalProperties.content = aprop;

	var ritems = new XMLList();

	var masters = {};
	for each (var id in aIds) {
		var itid;

		switch (id.type) {
		case "Occurrence":
		case "Exception":
			/* collect every master only once, but also collect every exception */
			if (!masters[id.uid]) {
				itid = <t:RecurringMasterItemId OccurrenceId={id.Id} xmlns:t={t}/>;
				ritems += itid;
				masters[id.uid] = true;
			}
			/* don't want Occurrences as items, only as master */
			if (id.type == "Occurrence") {
				break;
			}
			/* FALLTHROUGH */
		default:
			itid = <t:ItemId Id={id.Id} xmlns:t={t}/>;
			ritems += itid;
			break;
		}
	}

	req.m::ItemIds.content = ritems;

	return makeSoapMessage(req);
}

function makeTzSection(aType, aTz)
{
	var r = <t:{aType} xmlns:t={t}/>;

	var rrule = aTz.getFirstProperty("RRULE");

	var m = aTz.getFirstProperty("TZOFFSETTO").value.match(/^([+-]?)(\d\d)(\d\d)$/);
	var offs = cal.createDuration();
	offs.hours = m[2];
	offs.minutes = m[3];
	if (m[1] == '+') {
		offs.isNegative = true;
	}

	m = aTz.getFirstProperty("TZOFFSETFROM").value.match(/^([+-]?)(\d\d)(\d\d)$/);
	var offs_from = cal.createDuration();
	offs_from.hours = m[2];
	offs_from.minutes = m[3];
	if (m[1] == '+') {
		offs_from.isNegative = true;
	}

	r.t::Offset = offs.icalString;

	/* can't get parameters of RRULEs... have to do it manually :/ */
	if (rrule) {
	    var prop = {};
	    for each (let ps in rrule.value.split(';')) {
		    let m = ps.split('=');
		    prop[m[0]] = m[1];
	    }

	    var m = prop["BYDAY"].match(/^(-?\d)(..)$/);
	    r.t::RelativeYearlyRecurrence.t::DaysOfWeek = dayRevMap[m[2]];
	    r.t::RelativeYearlyRecurrence.t::DayOfWeekIndex = weekRevMap[m[1]];
	    r.t::RelativeYearlyRecurrence.t::Month = monthIdxMap[prop["BYMONTH"] - 1];
	}

	let start = aTz.getFirstProperty("DTSTART").valueAsDatetime.clone();
	start.addDuration(offs_from);
	offs.isNegative = !offs.isNegative;
	start.addDuration(offs);
	r.t::Time = start.toString().substr(11, 8);

	return (r);
}

function makeMeetingTimezone(aItem, e)
{
	var tz = aItem.startDate.timezone;

	if (tz.isUTC || tz.isFloating)
		return;

	var mtz = <t:MeetingTimeZone xmlns:t={t}/>;
	mtz.@TimeZoneName = tz.tzid;

	var tzcomp = tz.icalComponent;

	var dsttz = null;
	for (var comp = tzcomp.getFirstSubcomponent("DAYLIGHT");
	     comp;
	     comp = tzcomp.getNextSubcomponent("DAYLIGHT")) {
		if (!dsttz || dsttz.getFirstProperty("DTSTART").valueAsDatetime.compare(
				comp.getFirstProperty("DTSTART").valueAsDatetime) < 0) {
			dsttz = comp;
		}
	}
	var stdtz = null;
	for (var comp = tzcomp.getFirstSubcomponent("STANDARD");
	     comp;
	     comp = tzcomp.getNextSubcomponent("STANDARD")) {
		if (!stdtz || stdtz.getFirstProperty("DTSTART").valueAsDatetime.compare(
				comp.getFirstProperty("DTSTART").valueAsDatetime) < 0) {
			stdtz = comp;
		}
	}

	if (!stdtz) {
		return;
	}

	/* make the default offset zero */
	mtz.t::BaseOffset = cal.createDuration().icalString;

	mtz.appendChild(makeTzSection('Standard', stdtz));

	if (!dsttz) {
		mtz.t::BaseOffset = mtz.t::Standard.t::Offset.toString();
		delete mtz.t::Standard;
	} else {
		mtz.appendChild(makeTzSection('Daylight', dsttz));
	}

	e.appendChild(mtz);
}

function makeRecurrenceRule(aItem, e)
{
	if (!aItem.recurrenceInfo || aItem.parentItem != aItem) {
		return;
	}

	var rrule = null;
	for each (var ritem in aItem.recurrenceInfo.getRecurrenceItems({})) {
		if (calInstanceOf(ritem, Components.interfaces.calIRecurrenceRule)) {
			rrule = ritem;
			break;
		}
	}

	if (!rrule) {
		// XXX exception?
		return;
	}

	var r = <t:Recurrence xmlns:t={t}/>;

	/* can't get parameters of RRULEs... have to do it manually :/ */
	var prop = {};
	for each (let ps in rrule.icalProperty.value.split(';')) {
		let m = ps.split('=');
		prop[m[0]] = m[1];
	}

	var startDate = aItem.startDate.clone();
	startDate.isDate = true;

	prop["BYMONTHDAY"] = prop["BYMONTHDAY"] || startDate.day;
	prop["BYMONTH"] = prop["BYMONTH"] || (startDate.month + 1);

	switch (rrule.type) {
	case 'YEARLY':
		if (prop["BYDAY"]) {
			var m = prop["BYDAY"].match(/^(-?\d)(..)$/);
			r.t::RelativeYearlyRecurrence.t::DaysOfWeek = dayRevMap[m[2]];
			r.t::RelativeYearlyRecurrence.t::DayOfWeekIndex = weekRevMap[m[1]];
			r.t::RelativeYearlyRecurrence.t::Month = monthIdxMap[prop["BYMONTH"] - 1];
		} else {
			r.t::AbsoluteYearlyRecurrence.t::DayOfMonth = prop["BYMONTHDAY"];
			r.t::AbsoluteYearlyRecurrence.t::Month = monthIdxMap[prop["BYMONTH"] - 1];
		}
		break;
	case 'MONTHLY':
		if (prop["BYDAY"]) {
			r.t::RelativeMonthlyRecurrence.t::Interval = rrule.interval;
			var m = prop["BYDAY"].match(/^(-?\d)(..)$/);
			r.t::RelativeMonthlyRecurrence.t::DaysOfWeek = dayRevMap[m[2]];
			r.t::RelativeMonthlyRecurrence.t::DayOfWeekIndex = weekRevMap[m[1]];
		} else {
			r.t::AbsoluteMonthlyRecurrence.t::Interval = rrule.interval;
			r.t::AbsoluteMonthlyRecurrence.t::DayOfMonth = prop["BYMONTHDAY"];
		}
		break;
	case 'WEEKLY':
		r.t::WeeklyRecurrence.t::Interval = rrule.interval;
		var days = [];
		var daystr = prop["BYDAY"] || dayIdxMap[startDate.weekday];
		for each (let day in daystr.split(",")) {
			days.push(dayRevMap[day]);
		}
		r.t::WeeklyRecurrence.t::DaysOfWeek = days.join(' ');
		break;
	case 'DAILY':
		r.t::DailyRecurrence.t::Interval = rrule.interval;
		break;
	}

	if (rrule.isByCount && rrule.count != -1) {
		r.t::NumberedRecurrence.t::StartDate = cal.toRFC3339(startDate);
		r.t::NumberedRecurrence.t::NumberOfOccurrences = rrule.count;
	} else if (!rrule.isByCount && rrule.untilDate) {
		var endDate = rrule.untilDate.clone();
		endDate.isDate = true;
		r.t::EndDateRecurrence.t::StartDate = cal.toRFC3339(startDate);
		r.t::EndDateRecurrence.t::EndDate = cal.toRFC3339(endDate);
	} else {
		r.t::NoEndRecurrence.t::StartDate = cal.toRFC3339(startDate);
	}

	/* We won't write WKST/FirstDayOfWeek for now because it is Exchange 2010 and up */

	e.appendChild(r);
}

function makeDismissSnoozeState(aItem, e)
{
	let par = aItem.parentItem;
	let nextRem = cal.createDateTime("4501-01-01T00:00:00Z");

	let lastAck = par.alarmLastAck;
	let nextOcc;

	if (par.recurrenceInfo) {
		if (!lastAck) {
			/* no ack yet, get first occurrence */
			nextOcc = par.recurrenceInfo.getOccurrences(par.startDate, null, 1, {})[0];
		} else {
			nextOcc = par.recurrenceInfo.getNextOccurrence(lastAck);
		}
	} else {
		nextOcc = aItem;
	}

	let alarmDate = null;
	let alarm = aItem.getAlarms({})[0];
	if (nextOcc && alarm) {
		alarmDate = cal.alarms.calculateAlarmDate(nextOcc, alarm);
	}

	let lastSnooze = cal.createDateTime();
	if (par.getProperty('X-MOZ-SNOOZE-TIME')) {
		lastSnooze = cal.createDateTime(par.getProperty('X-MOZ-SNOOZE-TIME'));
	}

	let props = aItem.propertyEnumerator;
	while (props.hasMoreElements()) {
		let prop = props.getNext().QueryInterface(Components.interfaces.nsIProperty);
		if (prop.name.substr(0, 18) == 'X-MOZ-SNOOZE-TIME-') {
			let st = cal.createDateTime(prop.value);
			if (st.compare(lastSnooze) > 0) {
				lastSnooze = st;
			}
		}
	}

	if (lastSnooze.compare(cal.createDateTime()) > 0 &&	/* valid snooze */
	    alarmDate &&					/* have upcoming alarm */
	    lastSnooze.compare(alarmDate) < 0 &&		/* snooze before alarm */
	    (!lastAck || lastSnooze.compare(lastAck) > 0)) {	/* snozze after ack */
		nextRem = lastSnooze;
	} else if (alarmDate &&
		   (!lastAck || alarmDate.compare(lastAck) > 0)) {
		nextRem = alarmDate;
	}

	nextRem = nextRem.getInTimezone(cal.UTC());

	let eprop = <t:ExtendedProperty xmlns:t={t}/>;
	eprop.t::ExtendedFieldURI.@DistinguishedPropertySetId = "Common";
	eprop.t::ExtendedFieldURI.@PropertyId = "34144";
	eprop.t::ExtendedFieldURI.@PropertyType = "SystemTime";
	eprop.t::Value = cal.toRFC3339(nextRem);

	e.appendChild(eprop);
}

function makeXmlItem(aItem)
{
	var e = <t:CalendarItem xmlns:t={t} xmlns:m={m}/>;

	e.t::Subject = aItem.title;
	e.t::Body.@BodyType = "Text";
	e.t::Body = aItem.getProperty('DESCRIPTION') || "";

	var categories = aItem.getCategories({});
	for each (var category in categories) {
		e.t::Categories.list += <t:String xmlns:t={t}>{category}</t:String>;
	}

	var alarms = aItem.getAlarms({});
	if (alarms.length !== 0) {
		e.t::ReminderIsSet = 'true';

		// XXX only take the first for now
		let alarm = alarms[0];
		let alarmDate = cal.alarms.calculateAlarmDate(aItem, alarm);
		let offset = aItem.startDate.subtractDate(alarmDate);

		e.t::ReminderMinutesBeforeStart = offset.inSeconds / 60;
	} else {
		e.t::ReminderIsSet = 'false';
	}

	makeDismissSnoozeState(aItem, e);

	e.t::UID = aItem.id;

	e.t::Start = convDate(aItem.startDate);
	e.t::End = convDate(aItem.endDate);

	e.t::IsAllDayEvent = aItem.startDate.isDate;

	e.t::Location = aItem.getProperty("LOCATION") || "";

	var attendees = aItem.getAttendees({});
	for each (var attendee in attendees) {
		const attendeeStatus = {
			"NEEDS-ACTION"	: "Unknown",
			"TENTATIVE"	: "Tentative",
			"ACCEPTED"	: "Accept",
			"DECLINED"	: "Decline",
			null		: "Unknown",
		};

		var ae = <t:Attendee xmlns:t={t}/>;

		ae.t::Mailbox.t::Name = attendee.commonName;
		ae.t::Mailbox.t::EmailAddress = attendee.id.replace(/^mailto:/, '');
		ae.t::ResponseType = attendeeStatus[attendee.participationStatus];

		switch (attendee.role) {
		case "REQ-PARTICIPANT":
			e.t::RequiredAttendees.list += ae;
			break;
		case "OPT-PARTICIPANT":
			e.t::OptionalAttendees.list += ae;
			break;
		}
	}

	makeRecurrenceRule(aItem, e);
	makeMeetingTimezone(aItem, e);

	return e;
}

function makeCreateItem(aItem)
{
	var req = <m:CreateItem xmlns:m={m} xmlns:t={t}/>;
	req.@SendMeetingInvitations = sendMail(aItem);

	req.m::SavedItemFolderId.t::DistinguishedFolderId.@Id = "calendar";

	req.m::Items.content = makeXmlItem(aItem);

	return makeSoapMessage(req);
}

function makeItemId(aItem)
{
	var id;

	/* handle trivial single/master case */
	if (aItem.parentItem == aItem) {
		id = <t:ItemId xmlns:t={t}/>;
		id.@Id = aItem.getProperty("X-ITEMID");
		id.@ChangeKey = aItem.getProperty("SEQUENCE");
		return id;
	}

	var master = aItem.parentItem;

	/* if we got here, we're dealing with an occurrence (exception) */
	id = <t:OccurrenceItemId xmlns:t={t}/>;
	id.@RecurringMasterId = master.getProperty("X-ITEMID");
	id.@ChangeKey = master.getProperty("SEQUENCE");

	/*
	 * EWS can not deal with RecurrenceIds, but only with
	 * numerical indices to identify an occurrence.
	 *
	 * To determine our index, we will have to enumerate all
	 * regular occurrences up to our original date and count them.
	 */

	/* first find the recurrence rule */
	var rrule = null;
	for each (let ritem in master.recurrenceInfo.getRecurrenceItems({})) {
		if (calInstanceOf(ritem, Components.interfaces.calIRecurrenceRule)) {
			rrule = ritem;
			break;
		}
	}

	/*
	 * The end date needs to be after our original start date.  To be
	 * on the safe side, add the event duration.
	 */
	var enddate = aItem.recurrenceId.clone();
	enddate.addDuration(master.endDate.subtractDate(master.startDate));

	var occurrences = rrule.getOccurrences(master.startDate,
					       master.startDate,
					       enddate,
					       0,
					       {});

	id.@InstanceIndex = occurrences.length;

	return id;
}

function getDeletedOccurrences(aItem)
{
	var deleted = {};

	if (!aItem.recurrenceInfo) {
		return deleted;
	}

	for each (let ritem in aItem.recurrenceInfo.getRecurrenceItems({})) {
		if (ritem.isNegative) {
			var oc = aItem.recurrenceInfo.getOccurrenceFor(ritem.date);
			var rid = cal.toRFC3339(oc.recurrenceId);
			deleted[rid] = oc;
		}
	}

	return deleted;
}

function checkDeletedOccurrences(aNewItem, aOldItem)
{
	/*
	 * This function gets confused when the start date of the
	 * parent did change.  For now just assume that if the
	 * start date changed we won't delete occurrences at the same time.
	 */
	if (aNewItem.startDate.compare(aOldItem.startDate) != 0) {
		return null;
	}

	var newlist = getDeletedOccurrences(aNewItem);
	var oldlist = getDeletedOccurrences(aOldItem);

	var dellist = [];
	for (let del in newlist) {
		if (!oldlist[del]) {
			dellist.push(newlist[del]);
		}
	}

	if (dellist.length == 0) {
		return null;
	}

	var req = <m:DeleteItem xmlns:m={m}/>;
	req.@DeleteType = "HardDelete";
	req.@SendMeetingCancellations = sendMail(aOldItem);

	var ids = new XMLList();
	for each (let id in dellist) {
		ids += makeItemId(id);
	}
	req.m::ItemIds.content = ids;

	return req;
}

function makeUpdateOneItem(aNewItem, aOldItem, aFlags)
{
	var upd = <t:ItemChange xmlns:t={t}/>;
	upd.id = makeItemId(aOldItem);

	const noDelete = {
		'MeetingTimeZone'		: true,
		'ReminderMinutesBeforeStart'	: true
	};

	var ce = new XMLList();

	var oe = makeXmlItem(aOldItem);
	var ne = makeXmlItem(aNewItem);

	for each (var prop in oe) {
		if (ne[prop.name()].length() > 0 || noDelete[prop.localName()]) {
			continue;
		}

		var de = <t:DeleteItemField xmlns:t={t}/>;
		if (prop.localName() == "ExtendedProperty") {
			de.t::ExtendedFieldURI = prop.t::ExtendedFieldURI;
		} else {
			de.t::FieldURI.@FieldURI = fieldPathMap[prop.localName()] + ':' + prop.localName();
		}

		ce += de;
	}

	for each (var prop in ne) {
		if (oe.children().contains(prop)) {
			continue;
		}

		var se = <t:SetItemField xmlns:t={t}/>;
		if (prop.localName() == "ExtendedProperty") {
			se.t::ExtendedFieldURI = prop.t::ExtendedFieldURI;
		} else {
			se.t::FieldURI.@FieldURI = fieldPathMap[prop.localName()] + ':' + prop.localName();
		}
		se.t::CalendarItem.content = prop;

		ce += se;
	}

	upd.t::Updates.content = ce;

	if (!ce[0]) {
		return null;
	}

	/*
	 * If we change the recurrence, EWS will delete all exceptions.
	 * tell our caller that it will have to recreate them later.
	 */
	if (se..t::Recurrence[0]) {
		for each (let ritem in aNewItem.recurrenceInfo.getRecurrenceItems({})) {
			if (cal.calInstanceOf(ritem, Components.interfaces.calIRecurrenceDate)) {
				aFlags.createExceptions = true;
			}
		}
	}

	return upd;
}

function makeUpdateItem2(aItem, aUpdates)
{
	var req = <m:UpdateItem xmlns:m={m} xmlns:t={t}/>;
	req.@MessageDisposition = "SaveOnly";
	req.@ConflictResolution = "AutoResolve";
	req.@SendMeetingInvitationsOrCancellations = sendMail(aItem);

	req.m::ItemChanges.content = aUpdates;

	return req;
}

function makeUpdateItem(aNewItem, aOldItem, aFlags)
{
	var dreq = checkDeletedOccurrences(aNewItem, aOldItem);
	if (dreq) {
		return makeSoapMessage(dreq);
	}

	var upd = makeUpdateOneItem(aNewItem, aOldItem, aFlags);

	if (!upd) {
		return "";
	}

	var req = makeUpdateItem2(aNewItem, upd);

	return makeSoapMessage(req);
}

function makeDeleteItem(aItem)
{
	var req = <m:DeleteItem xmlns:m={m}/>;
	req.@DeleteType = "HardDelete";
	req.@SendMeetingCancellations = sendMail(aItem);

	req.m::ItemIds.id = makeItemId(aItem);

	return makeSoapMessage(req);
}

function makeMeetingResponse(aItem, aResp)
{
	var req = <m:CreateItem xmlns:m={m}/>;
	req.@SendMeetingInvitations = sendMail(aItem);
	req.@MessageDisposition = "SendAndSaveCopy";

	const responseMap = {
		// XXX
		"NEEDS-ACTION"	: <t:TentativelyAcceptItem xmlns:t={t}/>,
		"TENTATIVE"	: <t:AcceptItem xmlns:t={t}/>,

		"ACCEPTED"	: <t:AcceptItem xmlns:t={t}/>,
		"DECLINED"	: <t:DeclineItem xmlns:t={t}/>
	};

	var r = responseMap[aResp];

	if (!r) {
		return "";
	}
	r.t::ReferenceItemId.@Id = aItem.getProperty("X-ITEMID");
	r.t::ReferenceItemId.@ChangeKey = aItem.getProperty("SEQUENCE");

	req.m::Items.content = r;

	return makeSoapMessage(req);
}

function makeGetUserAvailability(aId, aStart, aEnd)
{
	var req = <m:GetUserAvailabilityRequest xmlns:m={m}/>;

	/* WTF really?  Just give me UTC. */
	req.t::TimeZone.t::Bias = 0;
	req.t::TimeZone.t::StandardTime.t::Bias = 0;
	req.t::TimeZone.t::StandardTime.t::Time = '00:00:00';
	req.t::TimeZone.t::StandardTime.t::DayOrder = 1;
	req.t::TimeZone.t::StandardTime.t::Month = 0;
	req.t::TimeZone.t::StandardTime.t::DayOfWeek = "Sunday";
	req.t::TimeZone.t::DaylightTime.content = req.t::TimeZone.t::StandardTime.children();

	req.m::MailboxDataArray.t::MailboxData.t::Email.t::Address = aId;
	req.m::MailboxDataArray.t::MailboxData.t::AttendeeType = 'Required';

	req.t::FreeBusyViewOptions.t::TimeWindow.t::StartTime = convDate(aStart);
	req.t::FreeBusyViewOptions.t::TimeWindow.t::EndTime = convDate(aEnd);

	req.t::FreeBusyViewOptions.t::MergedFreeBusyIntervalInMinutes = 15;

	return makeSoapMessage(req);
}
